from .core.cp import (CCRM, CPPM, CPRM, CounterPropagation,
                      CounterPropagationClassWiseRegModel,
                      CounterPropagationProtoModel, CounterPropagationRegModel)
from .core.initializer import (KMEI, NGRI, RELI, KMeans_Initializer,
                               NGRInitializer, ResponseLikeInitializer)
from .core.response_layer import NGR, REM, NeuralGasResponse, ResponseModel



from typing import Any, Callable, Optional

import torch
from prototorch.models.abstract import ProtoTorchBolt
from torch.nn import Linear, MSELoss
from torch.nn.parameter import Parameter

from counterpropagation.utils import accuracy


class CounterPropagation(ProtoTorchBolt):

    """

    Counterpropagation Model - Base Class

    Inherits:
        ProtoTorchBolt: ProtoTorchBolt to migrate into the Prototorch Framework
    
    Forward Flow:
        Input -> ResponseModel(Input) -> SupervisedModel(Response) -> Loss(SupervisedModel)

    Backward Flow:
                            |----> ResponseModel(Input)          
                            |    
                            |
        Loss(SupervisedModel)
                            |                            
                            |
                            |----> SupervisedModel(Response)

    Thus, ResponseModel must be differentiable by Version 1.0.1, however future updates 
    may include non-trainable Response Models.

    """

    def __init__( 
        self, 
        hparams,
        resmodel: Any,
        opsmodel: Optional[Any] = None,
        train_metrics: dict = dict(acc=accuracy),
        test_metrics: Optional[dict] = None,
        same_metrics: bool = True,
        **kwargs 
        ):

        """

        Args:
            hparams (dict): 
            hparams for the concatenated models
            
            resmodel (Any): 
            Response Model
            
            opsmodel (Optional[Any], None): 
            Supervised (Operation) Model (Layer), can be None for regression,
            in which case an perceptron like layer is created.
             Defaults to None.
            
            train_metrics (dict): 
            Metrics being used to monitor training performance in shape like:
            train_metrics = {"name_of_metric": callable_metric} 
             Defaults to accuracy.
            
            test_metrics (Optional[dict], None): 
            Metrics being used to monitor test performance. 
             Defaults to None.
            
            same_metrics (bool): 
            If True, test_metrics will equal train_metrics. If False a dict needs to be
            given to test_metrics with the corresponding metrics. 
             Defaults to True.

        """

        super(CounterPropagation, self).__init__(hparams, **kwargs)

        if opsmodel is None:
            self.res_layer = resmodel(hparams, prototypes_initializer=self.hparams.response_initializer, **kwargs)
            self.ops_layer = opsmodel
        else:
            self.res_layer = resmodel(hparams, prototypes_initializer=self.hparams.response_initializer, **kwargs)
            self.ops_layer = opsmodel(hparams, prototypes_initializer=self.hparams.operation_initializer, **kwargs)
        
        ## allow for customizable metrics
        ## and add test to every key in
        ## test metrics if test and train
        ## metrics are equal
        if same_metrics:
            met_copy = train_metrics.copy()
            for k in train_metrics.keys():
                new_k = "test_" + k
                met_copy[new_k] = train_metrics[k]
                del met_copy[k]
        else:
            assert test_metrics is not None, "Provide test_metrics if same_metrics is False"
            met_copy = test_metrics.copy()
            del test_metrics
            
        self.train_metrics = train_metrics
        self.test_metrics = met_copy

    def forward(self, batch: Any):
        """

        To modify the forward a custom end_step function can be created,
        which depends on the batch (necessary for class-wise regression models)

        Args:
            batch (Any): 
            Used for end_step

        Returns:
            torch.Tensor: the predictions.

        """

        x, y = batch
        response = self.res_layer(x)
        predictions = self.end_step(response, y)
        return predictions
    
    def _log_metric(self, fun: Callable, batch, tag: str):

        """

        Allows to log any custom metric which depends
        on predictions and targets.
        
        Args:
            fun (Callable): 
            A custom metric m, s.t. m(predictions, targets)

            batch (Any): 
            Current batch

            tag (str): 
            Name of the metric

        """

        _, y = batch
        with torch.no_grad():
            predictions = self.forward(batch)
            result = fun(predictions, y)
        
        self.log( 
            tag,
            result,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
    
    def predict(self, batch):
        return self.forward(batch)
    
    def training_step(self, batch, batch_idx):
        _, train_loss = self.shared_step(batch, batch_idx)
        self.log("train_loss", train_loss)
        for k, v in self.train_metrics.items():
            self._log_metric(v, batch, tag=k)
        return train_loss
    
    def test_step(self, batch, batch_idx):
        _, test_loss = self.shared_step(batch, batch_idx)
        for k, v in self.test_metrics.items():
            self._log_metric(v, batch, tag=k)
        return test_loss

class CounterPropagationProtoModel(CounterPropagation):

    """

    Prototype-based Classification CP-Model, i.e. opsmodel is a
    prototype-based / distance-based Model

    """

    def __init__(self, hparams, **kwargs):
        """

        Args:
            hparams (dict): 
            as in base class CounterPropagation

        Raises:
            ValueError: When opsmodel is None.

        """

        super(CounterPropagationProtoModel, self).__init__(hparams, **kwargs)
        if self.ops_layer is None:
            raise ValueError("Supervised Layer in Prototype Model cannot be None")
        
        self.omega_available = False
        if hasattr(self.ops_layer, "_omega"):
            self.ops_omega = Parameter(self.ops_layer._omega)
            self.omega_available = True
        
        self.ops_dl = self.ops_layer.distance_layer
        self.ops_protos = Parameter(self.ops_layer.prototypes)

    def shared_step(self, batch, batch_idx, opt_idx=None):
        x, y = batch
        response = self.res_layer(x)
        
        if self.omega_available:
            distances = self.ops_dl(response, self.ops_protos, self.ops_omega)
        else:
            distances = self.ops_dl(response, self.ops_protos)
        
        _, ops_label = self.ops_layer.proto_layer()
        loss = self.ops_layer.loss(distances, y, ops_label)
        return distances, loss
    
    def end_step(self, batch_reponse, targets):
        response = batch_reponse
        return self.ops_layer.predict(response)

class CounterPropagationRegModel(CounterPropagation):

    """

    Regression based CP-Model (for classification or regression).
    opsmodel is allowed to be None in which case a torch.nn.Linear
    instance is created.

    """

    def __init__(self, hparams, loss: Callable = MSELoss(reduction="sum"), **kwargs):

        """

        Regression-based CP-Model, allows opsmodel to be None.

        Args:
            hparams (dict): 
            as in base class CounterPropagation
        
            loss (Callable, torch.nn.MSELoss):
            torch.Module like or any Callable (which is differentiable).
            Default loss is torch.nn.MSELoss.

        """

        super(CounterPropagationRegModel, self).__init__(hparams, **kwargs)

        self.idim = self.hparams.res_dim
        if self.ops_layer is None:
            self.ops_layer = Linear(self.idim, 1)
        
        self.loss = loss

    def shared_step(self, batch, batch_idx, opt_idx=None):
        x, y = batch
        response = self.res_layer(x)
        preds = self.ops_layer(response).squeeze()
        loss = self.loss(preds, 1. * y)
        return preds, loss

    def end_step(self, batch_reponse, targets):
        response = batch_reponse
        return self.ops_layer(response).squeeze()

class CounterPropagationClassWiseRegModel(CounterPropagationRegModel):

    """

    Class-wise Regression based CP-Model, i.e. for each class there is
    a responsible perceptron.

    """

    def __init__(self, hparams, **kwargs):

        """

        When opsmodel is None a linear-like layer is created by using
        res_dim and distribution provided in hparams.
        distribution is assumed to be of the shape dict(num_classes=n, per_class=p)


        """

        super(CounterPropagationClassWiseRegModel, self).__init__(hparams, **kwargs)

        nc, npc = self.hparams.distribution["num_classes"], self.hparams.distribution["per_class"]
        out_dim = nc * npc
        
        self.ops_layer = torch.nn.Linear(self.idim, out_dim)
        self.perceptron_label = torch.LongTensor([i for i in range(nc) for j in range(npc)])
    
    def _class_filter(self, y):
        plabs = self.perceptron_label
        if y.ndim == 1:
            return y.unsqueeze(-1) == plabs
        elif y.ndim == 2:
            return torch.max(y, -1).indices.unsqueeze(-1) == plabs
        else:
            raise TypeError(f"Only supported up to 2D labels, but got {y.ndim}D")

    def shared_step(self, batch, batch_idx, opt_idx=None):
        x, y = batch
        response = self.res_layer(x)
        raw_preds = self.ops_layer(response).squeeze()
        filter_ = self._class_filter(y)
        preds = raw_preds[filter_]
        loss = self.loss(preds, 1.*y)
        return preds, loss
    
    def end_step(self, batch_reponse, targets):
        response, y = batch_reponse, targets
        raw_preds = self.ops_layer(response)
        preds = raw_preds[self._class_filter(y)]
        return preds
        
    


CPPM = CounterPropagationProtoModel
CPRM = CounterPropagationRegModel
CCRM = CounterPropagationClassWiseRegModel



import torch
from prototorch.models.vis import Vis2DAbstract


class VisCP2D(Vis2DAbstract):

    """

    Visualization during Training based on Vis2DAbstract by Prototorch.
    
    Adapted for Counterpropagation to visualize the sup_layer in 
    the response-space (Only for Prototype-based sup_layer CP).
    Takes the first 2 dimensions for visualization.
    
    """

    def __init__(self, *args, **kwargs):
        super(VisCP2D, self).__init__(*args, **kwargs)

    def visualize(self, pl_module):
        with torch.no_grad():
            _, plabel = pl_module.ops_layer.proto_layer()
            protos = pl_module.ops_protos
            response = pl_module.res_layer(self.x_train)
            predictions = pl_module.predict((self.x_train, self.y_train))
            ax = self.setup_ax()
            self.plot_data(ax, response, predictions)
            self.plot_protos(ax, protos, plabel)
        



from typing import Callable

import torch
from prototorch.core.distances import squared_euclidean_distance
from prototorch.core.initializers import AbstractComponentsInitializer
from torch_kmeans import KMeans

from counterpropagation.utils import neighbourhood_fun


class KMeans_Initializer(AbstractComponentsInitializer):
    ## uses https://pypi.org/project/torch-kmeans/
    def __init__(
            self,
            hparams,
            data: torch.Tensor,
            noise: float = 0.0,
            transform: Callable = torch.nn.Identity(),
    ):
        self.data = data
        self.noise = noise
        self.transform = transform
        self.model = KMeans(**hparams)
    
    def generate_end_hook(self, center):
        drift = torch.rand_like(center) * self.noise
        components = self.transform(center + drift)
        return components

    def generate(self, num_components=None):
        model_result = self.model(self.data.unsqueeze(0))
        centers = model_result.centers.squeeze()
        components = self.generate_end_hook(centers)
        return components


class ResponseLikeInitializer(AbstractComponentsInitializer):

    """

    Base Class to create components in the response-space.
    A corresponding generate funvtion needs to be created 
    for the respective response-space.

    For good initializations x should be equal or close 
    to the response space.
    """

    def __init__(self, x, response_protos, noise: float = 0.0, transform: Callable = torch.nn.Identity()):
        super(ResponseLikeInitializer, self).__init__()
        self.x = x
        self.res_protos = response_protos
        self.noise = noise
        self.transform = transform

    def generate_end_hook(self, x):
        drift = torch.rand_like(x) * self.noise
        components = self.transform(x + drift)
        return components
    
    def generate(self, num_components):
        raise NotImplementedError

class NGRInitializer(ResponseLikeInitializer):

    """

    Creates components in a neural gas like response-space.

    """

    def __init__(self, lmbda: float = 1., **kwargs):
        super(NGRInitializer, self).__init__(**kwargs)

        self.lmbda = lmbda
    
    def generate(self, num_components):
        dists = squared_euclidean_distance(self.x, self.res_protos)
        response = neighbourhood_fun(dists, self.lmbda) * dists
        protos = torch.empty((num_components, response.shape[1]))
        for i in range(num_components):
            inds = torch.LongTensor(num_components).random_(0, len(self.x))
            protos[i] = torch.mean(response[inds], 0)
        return self.generate_end_hook(protos)
        

RELI = ResponseLikeInitializer
NGRI = NGRInitializer
KMEI = KMeans_Initializer



from typing import Optional

import torch
from pytorch_lightning.callbacks import Callback

NG_DECAY = dict( 
    exp="Exponential decay of the form lambda ^ decay_constant",
    mul="Multiplicative decay of the form lambda * decay_constant"
)
DESCRIPTION = "Decay Options are:\n" + "\n".join([f"{k} - {v}" for (k, v) in NG_DECAY.items()])

class NGCallback(Callback):
    """

    Callback for the schedule of the neighborhood-range lambda in Neural Gas

    """
    def __init__(self, lmbda_start: Optional[float] = 1., gamma: float = 0.99, decay_type: str = "exp"):
        super(NGCallback, self).__init__()

        if lmbda_start is None:
            self.lmbda = lmbda_start
        else:
            self.lmbda_start = torch.Tensor([lmbda_start])
    
        assert decay_type in NG_DECAY, print(DESCRIPTION)
        self.decay_type = decay_type
        decay_funs = dict( 
            exp=self.exp_decay,
            mul=self.mul_decay,
        )
        self.decay_fun = decay_funs[self.decay_type]

        assert gamma <= 1. and gamma > 0., "gamma must be less or equal 1. and positive."
        self.gamma = gamma
    
    def exp_decay(self, l):
        return l ** self.gamma

    def mul_decay(self, l):
        return l * self.gamma
    
    def on_train_epoch_end(self, trainer, pl_module) -> None:
        sd = pl_module.res_layer.state_dict()
        if self.lmbda_start is None:
            l = sd["lmbda"]
        else:
            l = self.lmbda_start
        
        sd["lmbda"] = self.decay_fun(l)
        pl_module.res_layer.load_state_dict(sd)





import torch

from .callbacks import NGCallback
from .evalmeasure import accuracy, err10, r_squared
from .visualization import VisCP2D

rank_fun = lambda x: torch.argsort(torch.argsort(x, 1), 1)
neighbourhood_fun = lambda x, l: torch.exp(- rank_fun(x) / l)


import torch
from prototorch.core.distances import euclidean_distance
from prototorch.models.abstract import ProtoTorchBolt
from prototorch.nn.wrappers import LambdaLayer
from torch.nn import Parameter

from counterpropagation.utils import neighbourhood_fun


class ResponseModel(ProtoTorchBolt):

    """

    Base Class to create new models used for a Response in CP framework

    Inherits:
        ProtoTorchBolt: Base Class from Prototorch, since
                        responses are assumed to have no
                        other supervision then by the 
                        subsequent supervised layer.
                        
                        Changed to ProtoTorchBolt,
                        since prototypes in reslayer
                        were not detected as leaf in 
                        grad-tree.

    """

    def __init__(self, hparams, **kwargs):
        super(ResponseModel, self).__init__(hparams, **kwargs)
        distance_fn = kwargs.get("response_distance", euclidean_distance)
        self.distance_layer = LambdaLayer(distance_fn)

        proto_init = kwargs.get("prototypes_initializer", None)
        if proto_init is not None:
            protos = proto_init.generate(self.hparams.n_resprotos)
            self.register_parameter("response_components", Parameter(protos))
    
    @property
    def prototypes(self):
        return self.response_components
        
    def compute_distances(self, x):
        return self.distance_layer(x, self.prototypes)
    
    def forward(self, x):
        return self.response(x)

    def response(self, x):
        raise NotImplementedError

class NeuralGasResponse(ResponseModel):

    """

    Neural Gas like Response

    Inherits:
        ResponseModel: Base Class

    """

    def __init__(self, hparams, lmbda: float = 1., **kwargs):
        super(NeuralGasResponse, self).__init__(hparams, **kwargs)
        self.lmbda = Parameter(torch.tensor(lmbda), requires_grad=False)
    
    def response(self, x):
        distances = self.compute_distances(x)
        responses = neighbourhood_fun(distances, self.lmbda) * distances
        return responses

REM = ResponseModel
NGR = NeuralGasResponse




from setuptools import find_packages, setup

PROJECT_URL = "https://github.com/rmschubert/Counterpropagation-Prototorch"
DOWNLOAD_URL = "https://github.com/rmschubert/Counterpropagation-Prototorch.git"

with open("README.md", "r") as fh:
    long_description = fh.read()

INSTALL_REQUIRES = [
    "prototorch>=0.7.4",
    "prototorch_models>=0.5.4",
    "pytorch-lightning",
    "scipy",
    "torch-kmeans"
]


setup(
    name="counterprop_prototorch",
    version="1.0.3",
    description="Counter Propagation based on "
    "Prototorch. ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ronny Schubert",
    author_email="trebuhcsynnor@gmail.com",
    url=PROJECT_URL,
    download_url=DOWNLOAD_URL,
    license="MIT",
    python_requires=">=3.10",
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
    ],
    extras_require={ 
        "dev": ["twine>=4.0.2"],
    },
    packages=find_packages(),
    zip_safe=False,
)


import torch
from scipy.stats import linregress


def r_squared(p, y):
    _, _, r, _, _ = linregress(p.numpy(), y.numpy())
    return r**2

def err10(p, y):
    err = torch.abs(p - y)
    err10 = torch.sum(torch.where(err <= 0.1 * y, 1, 0)) / len(y)
    return err10

def accuracy(p, y):
    pref = 1 / (len(y))
    return pref * torch.sum(p.int() == y.int())
